mirror of
https://github.com/NVIDIA/cuda-samples.git
synced 2026-10-13 00:08:26 +08:00
CUDA 13.2 samples update (#432)
- Added Python samples for CUDA Python 1.0 release - Renamed top-level `Samples` directory to `cpp` to accommodate Python samples.
This commit is contained in:
18
cpp/5_Domain_Specific/marchingCubes/.vscode/c_cpp_properties.json
vendored
Normal file
18
cpp/5_Domain_Specific/marchingCubes/.vscode/c_cpp_properties.json
vendored
Normal file
@@ -0,0 +1,18 @@
|
||||
{
|
||||
"configurations": [
|
||||
{
|
||||
"name": "Linux",
|
||||
"includePath": [
|
||||
"${workspaceFolder}/**",
|
||||
"${workspaceFolder}/../../../Common"
|
||||
],
|
||||
"defines": [],
|
||||
"compilerPath": "/usr/local/cuda/bin/nvcc",
|
||||
"cStandard": "gnu17",
|
||||
"cppStandard": "gnu++14",
|
||||
"intelliSenseMode": "linux-gcc-x64",
|
||||
"configurationProvider": "ms-vscode.makefile-tools"
|
||||
}
|
||||
],
|
||||
"version": 4
|
||||
}
|
||||
7
cpp/5_Domain_Specific/marchingCubes/.vscode/extensions.json
vendored
Normal file
7
cpp/5_Domain_Specific/marchingCubes/.vscode/extensions.json
vendored
Normal file
@@ -0,0 +1,7 @@
|
||||
{
|
||||
"recommendations": [
|
||||
"nvidia.nsight-vscode-edition",
|
||||
"ms-vscode.cpptools",
|
||||
"ms-vscode.makefile-tools"
|
||||
]
|
||||
}
|
||||
90
cpp/5_Domain_Specific/marchingCubes/CMakeLists.txt
Normal file
90
cpp/5_Domain_Specific/marchingCubes/CMakeLists.txt
Normal file
@@ -0,0 +1,90 @@
|
||||
cmake_minimum_required(VERSION 3.20)
|
||||
|
||||
list(APPEND CMAKE_MODULE_PATH "${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/Modules")
|
||||
|
||||
project(marchingCubes LANGUAGES C CXX CUDA)
|
||||
|
||||
find_package(CUDAToolkit REQUIRED)
|
||||
|
||||
set(CMAKE_POSITION_INDEPENDENT_CODE ON)
|
||||
|
||||
set(CMAKE_CUDA_ARCHITECTURES 75 80 86 87 89 90 100 110 120)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -Wno-deprecated-gpu-targets")
|
||||
if(ENABLE_CUDA_DEBUG)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -G") # enable cuda-gdb (may significantly affect performance on some targets)
|
||||
else()
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -lineinfo") # add line information to all builds for debug tools (exclusive to -G option)
|
||||
endif()
|
||||
|
||||
# Include directories and libraries
|
||||
include_directories(../../../Common)
|
||||
|
||||
if(WIN32)
|
||||
set(PC_GLUT_INCLUDE_DIRS "${CMAKE_CURRENT_SOURCE_DIR}/../../../Common")
|
||||
set(PC_GLUT_LIBRARY_DIRS "${CMAKE_CURRENT_SOURCE_DIR}/../../../Common/lib/x64")
|
||||
endif()
|
||||
|
||||
find_package(OpenGL)
|
||||
find_package(GLUT)
|
||||
|
||||
# Source file
|
||||
if(${OpenGL_FOUND})
|
||||
if (${GLUT_FOUND})
|
||||
# Add target for marchingCubes
|
||||
add_executable(marchingCubes marchingCubes_kernel.cu marchingCubes.cpp)
|
||||
|
||||
target_compile_options(marchingCubes PRIVATE $<$<COMPILE_LANGUAGE:CUDA>:--extended-lambda>)
|
||||
|
||||
target_compile_features(marchingCubes PRIVATE cxx_std_17 cuda_std_17)
|
||||
|
||||
set_target_properties(marchingCubes PROPERTIES CUDA_SEPARABLE_COMPILATION ON)
|
||||
|
||||
target_include_directories(marchingCubes PUBLIC
|
||||
${OPENGL_INCLUDE_DIR}
|
||||
${CUDAToolkit_INCLUDE_DIRS}
|
||||
${GLUT_INCLUDE_DIRS}
|
||||
)
|
||||
|
||||
target_link_libraries(marchingCubes
|
||||
${OPENGL_LIBRARIES}
|
||||
${GLUT_LIBRARIES}
|
||||
)
|
||||
|
||||
# Copy data files to the output directory
|
||||
add_custom_command(TARGET marchingCubes POST_BUILD
|
||||
COMMAND ${CMAKE_COMMAND} -E copy_directory
|
||||
${CMAKE_CURRENT_SOURCE_DIR}/data
|
||||
${CMAKE_CURRENT_BINARY_DIR}/data
|
||||
)
|
||||
|
||||
if(WIN32)
|
||||
target_link_libraries(marchingCubes
|
||||
${PC_GLUT_LIBRARY_DIRS}/freeglut.lib
|
||||
${PC_GLUT_LIBRARY_DIRS}/glew64.lib
|
||||
)
|
||||
|
||||
add_custom_command(TARGET marchingCubes
|
||||
POST_BUILD
|
||||
COMMAND ${CMAKE_COMMAND} -E copy
|
||||
${CMAKE_CURRENT_SOURCE_DIR}/../../../bin/win64/$<CONFIGURATION>/freeglut.dll
|
||||
${CMAKE_CURRENT_BINARY_DIR}/$<CONFIGURATION>
|
||||
)
|
||||
|
||||
add_custom_command(TARGET marchingCubes
|
||||
POST_BUILD
|
||||
COMMAND ${CMAKE_COMMAND} -E copy
|
||||
${CMAKE_CURRENT_SOURCE_DIR}/../../../bin/win64/$<CONFIGURATION>/glew64.dll
|
||||
${CMAKE_CURRENT_BINARY_DIR}/$<CONFIGURATION>
|
||||
)
|
||||
endif()
|
||||
|
||||
else()
|
||||
message(STATUS "GLUT not found - will not build sample 'marchingCubes'")
|
||||
endif()
|
||||
else()
|
||||
message(STATUS "OpenGL not found - will not build sample 'marchingCubes'")
|
||||
endif()
|
||||
|
||||
# Include installation configuration
|
||||
include(${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/InstallSamples.cmake)
|
||||
setup_samples_install()
|
||||
36
cpp/5_Domain_Specific/marchingCubes/README.md
Normal file
36
cpp/5_Domain_Specific/marchingCubes/README.md
Normal file
@@ -0,0 +1,36 @@
|
||||
# marchingCubes - Marching Cubes Isosurfaces
|
||||
|
||||
## Description
|
||||
|
||||
This sample extracts a geometric isosurface from a volume dataset using the marching cubes algorithm. It uses the scan (prefix sum) function from the Thrust library to perform stream compaction.
|
||||
|
||||
## Key Concepts
|
||||
|
||||
OpenGL Graphics Interop, Vertex Buffers, 3D Graphics, Physically Based Simulation
|
||||
|
||||
## Supported SM Architectures
|
||||
|
||||
[SM 5.0 ](https://developer.nvidia.com/cuda-gpus) [SM 5.2 ](https://developer.nvidia.com/cuda-gpus) [SM 5.3 ](https://developer.nvidia.com/cuda-gpus) [SM 6.0 ](https://developer.nvidia.com/cuda-gpus) [SM 6.1 ](https://developer.nvidia.com/cuda-gpus) [SM 7.0 ](https://developer.nvidia.com/cuda-gpus) [SM 7.2 ](https://developer.nvidia.com/cuda-gpus) [SM 7.5 ](https://developer.nvidia.com/cuda-gpus) [SM 8.0 ](https://developer.nvidia.com/cuda-gpus) [SM 8.6 ](https://developer.nvidia.com/cuda-gpus) [SM 8.7 ](https://developer.nvidia.com/cuda-gpus) [SM 8.9 ](https://developer.nvidia.com/cuda-gpus) [SM 9.0 ](https://developer.nvidia.com/cuda-gpus)
|
||||
|
||||
## Supported OSes
|
||||
|
||||
Linux, Windows
|
||||
|
||||
## Supported CPU Architecture
|
||||
|
||||
x86_64, armv7l
|
||||
|
||||
## CUDA APIs involved
|
||||
|
||||
### [CUDA Runtime API](http://docs.nvidia.com/cuda/cuda-runtime-api/index.html)
|
||||
cudaGLUnmapBufferObject, cudaGraphicsUnmapResources, cudaCreateChannelDesc, cudaMemcpy, cudaFree, cudaGLMapBufferObject, cudaGraphicsResourceGetMappedPointer, cudaGraphicsMapResources, cudaDestroyTextureObject, cudaGLUnregisterBufferObject, cudaCreateTextureObject, cudaGLRegisterBufferObject, cudaGraphicsUnregisterResource, cudaMalloc, cudaGraphicsGLRegisterBuffer
|
||||
|
||||
## Dependencies needed to build/run
|
||||
[X11](../../../README.md#x11), [OpenGL](../../../README.md#opengl), [Freeglut](../../../README.md#freeglut), [GLEW](../../../README.md#glew)
|
||||
|
||||
## Prerequisites
|
||||
|
||||
Download and install the [CUDA Toolkit](https://developer.nvidia.com/cuda-downloads) for your corresponding platform.
|
||||
Make sure the dependencies mentioned in [Dependencies]() section above are installed.
|
||||
|
||||
## References (for more details)
|
||||
BIN
cpp/5_Domain_Specific/marchingCubes/data/Bucky.raw
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/data/Bucky.raw
Normal file
Binary file not shown.
BIN
cpp/5_Domain_Specific/marchingCubes/data/compVoxelArray.bin
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/data/compVoxelArray.bin
Normal file
Binary file not shown.
BIN
cpp/5_Domain_Specific/marchingCubes/data/normalArray.bin
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/data/normalArray.bin
Normal file
Binary file not shown.
BIN
cpp/5_Domain_Specific/marchingCubes/data/posArray.bin
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/data/posArray.bin
Normal file
Binary file not shown.
File diff suppressed because one or more lines are too long
49
cpp/5_Domain_Specific/marchingCubes/defines.h
Normal file
49
cpp/5_Domain_Specific/marchingCubes/defines.h
Normal file
@@ -0,0 +1,49 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#ifndef _DEFINES_H_
|
||||
#define _DEFINES_H_
|
||||
|
||||
typedef unsigned int uint;
|
||||
typedef unsigned char uchar;
|
||||
|
||||
// if SAMPLE_VOLUME is 0, an implicit dataset is generated. If 1, a voxelized
|
||||
// dataset is loaded from file
|
||||
#define SAMPLE_VOLUME 1
|
||||
|
||||
// Using shared to store computed vertices and normals during triangle
|
||||
// generation
|
||||
// improves performance
|
||||
#define USE_SHARED 1
|
||||
|
||||
// The number of threads to use for triangle generation (limited by shared
|
||||
// memory size)
|
||||
#define NTHREADS 32
|
||||
|
||||
#define SKIP_EMPTY_VOXELS 1
|
||||
|
||||
#endif
|
||||
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_lg.png
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_lg.png
Normal file
Binary file not shown.
|
After Width: | Height: | Size: 52 KiB |
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_md.png
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_md.png
Normal file
Binary file not shown.
|
After Width: | Height: | Size: 51 KiB |
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_sm.png
Normal file
BIN
cpp/5_Domain_Specific/marchingCubes/doc/screenshot_sm.png
Normal file
Binary file not shown.
|
After Width: | Height: | Size: 21 KiB |
1039
cpp/5_Domain_Specific/marchingCubes/marchingCubes.cpp
Normal file
1039
cpp/5_Domain_Specific/marchingCubes/marchingCubes.cpp
Normal file
File diff suppressed because it is too large
Load Diff
747
cpp/5_Domain_Specific/marchingCubes/marchingCubes_kernel.cu
Normal file
747
cpp/5_Domain_Specific/marchingCubes/marchingCubes_kernel.cu
Normal file
@@ -0,0 +1,747 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#ifndef _MARCHING_CUBES_KERNEL_CU_
|
||||
#define _MARCHING_CUBES_KERNEL_CU_
|
||||
|
||||
#include <cuda_runtime_api.h>
|
||||
#include <helper_cuda.h> // includes for helper CUDA functions
|
||||
#include <helper_math.h>
|
||||
#include <stdio.h>
|
||||
#include <string.h>
|
||||
#include <thrust/device_vector.h>
|
||||
#include <thrust/scan.h>
|
||||
|
||||
#include "defines.h"
|
||||
#include "tables.h"
|
||||
|
||||
// textures containing look-up tables
|
||||
cudaTextureObject_t triTex;
|
||||
cudaTextureObject_t numVertsTex;
|
||||
|
||||
// volume data
|
||||
cudaTextureObject_t volumeTex;
|
||||
|
||||
extern "C" void allocateTextures(uint **d_edgeTable, uint **d_triTable, uint **d_numVertsTable)
|
||||
{
|
||||
checkCudaErrors(cudaMalloc((void **)d_edgeTable, 256 * sizeof(uint)));
|
||||
checkCudaErrors(cudaMemcpy((void *)*d_edgeTable, (void *)edgeTable, 256 * sizeof(uint), cudaMemcpyHostToDevice));
|
||||
cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindUnsigned);
|
||||
|
||||
checkCudaErrors(cudaMalloc((void **)d_triTable, 256 * 16 * sizeof(uint)));
|
||||
checkCudaErrors(cudaMemcpy((void *)*d_triTable, (void *)triTable, 256 * 16 * sizeof(uint), cudaMemcpyHostToDevice));
|
||||
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypeLinear;
|
||||
texRes.res.linear.devPtr = *d_triTable;
|
||||
texRes.res.linear.sizeInBytes = 256 * 16 * sizeof(uint);
|
||||
texRes.res.linear.desc = channelDesc;
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = false;
|
||||
texDescr.filterMode = cudaFilterModePoint;
|
||||
texDescr.addressMode[0] = cudaAddressModeClamp;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&triTex, &texRes, &texDescr, NULL));
|
||||
|
||||
checkCudaErrors(cudaMalloc((void **)d_numVertsTable, 256 * sizeof(uint)));
|
||||
checkCudaErrors(
|
||||
cudaMemcpy((void *)*d_numVertsTable, (void *)numVertsTable, 256 * sizeof(uint), cudaMemcpyHostToDevice));
|
||||
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypeLinear;
|
||||
texRes.res.linear.devPtr = *d_numVertsTable;
|
||||
texRes.res.linear.sizeInBytes = 256 * sizeof(uint);
|
||||
texRes.res.linear.desc = channelDesc;
|
||||
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = false;
|
||||
texDescr.filterMode = cudaFilterModePoint;
|
||||
texDescr.addressMode[0] = cudaAddressModeClamp;
|
||||
texDescr.readMode = cudaReadModeElementType;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&numVertsTex, &texRes, &texDescr, NULL));
|
||||
}
|
||||
|
||||
extern "C" void createVolumeTexture(uchar *d_volume, size_t buffSize)
|
||||
{
|
||||
cudaResourceDesc texRes;
|
||||
memset(&texRes, 0, sizeof(cudaResourceDesc));
|
||||
|
||||
texRes.resType = cudaResourceTypeLinear;
|
||||
texRes.res.linear.devPtr = d_volume;
|
||||
texRes.res.linear.sizeInBytes = buffSize;
|
||||
texRes.res.linear.desc = cudaCreateChannelDesc(8, 0, 0, 0, cudaChannelFormatKindUnsigned);
|
||||
|
||||
cudaTextureDesc texDescr;
|
||||
memset(&texDescr, 0, sizeof(cudaTextureDesc));
|
||||
|
||||
texDescr.normalizedCoords = false;
|
||||
texDescr.filterMode = cudaFilterModePoint;
|
||||
texDescr.addressMode[0] = cudaAddressModeClamp;
|
||||
texDescr.readMode = cudaReadModeNormalizedFloat;
|
||||
|
||||
checkCudaErrors(cudaCreateTextureObject(&volumeTex, &texRes, &texDescr, NULL));
|
||||
}
|
||||
|
||||
extern "C" void destroyAllTextureObjects()
|
||||
{
|
||||
checkCudaErrors(cudaDestroyTextureObject(triTex));
|
||||
checkCudaErrors(cudaDestroyTextureObject(numVertsTex));
|
||||
checkCudaErrors(cudaDestroyTextureObject(volumeTex));
|
||||
}
|
||||
|
||||
// an interesting field function
|
||||
__device__ float tangle(float x, float y, float z)
|
||||
{
|
||||
x *= 3.0f;
|
||||
y *= 3.0f;
|
||||
z *= 3.0f;
|
||||
return (x * x * x * x - 5.0f * x * x + y * y * y * y - 5.0f * y * y + z * z * z * z - 5.0f * z * z + 11.8f) * 0.2f
|
||||
+ 0.5f;
|
||||
}
|
||||
|
||||
// evaluate field function at point
|
||||
__device__ float fieldFunc(float3 p) { return tangle(p.x, p.y, p.z); }
|
||||
|
||||
// evaluate field function at a point
|
||||
// returns value and gradient in float4
|
||||
__device__ float4 fieldFunc4(float3 p)
|
||||
{
|
||||
float v = tangle(p.x, p.y, p.z);
|
||||
const float d = 0.001f;
|
||||
float dx = tangle(p.x + d, p.y, p.z) - v;
|
||||
float dy = tangle(p.x, p.y + d, p.z) - v;
|
||||
float dz = tangle(p.x, p.y, p.z + d) - v;
|
||||
return make_float4(dx, dy, dz, v);
|
||||
}
|
||||
|
||||
// sample volume data set at a point
|
||||
__device__ float sampleVolume(cudaTextureObject_t volumeTex, uchar *data, uint3 p, uint3 gridSize)
|
||||
{
|
||||
p.x = min(p.x, gridSize.x - 1);
|
||||
p.y = min(p.y, gridSize.y - 1);
|
||||
p.z = min(p.z, gridSize.z - 1);
|
||||
uint i = (p.z * gridSize.x * gridSize.y) + (p.y * gridSize.x) + p.x;
|
||||
// return (float) data[i] / 255.0f;
|
||||
return tex1Dfetch<float>(volumeTex, i);
|
||||
}
|
||||
|
||||
// compute position in 3d grid from 1d index
|
||||
// only works for power of 2 sizes
|
||||
__device__ uint3 calcGridPos(uint i, uint3 gridSizeShift, uint3 gridSizeMask)
|
||||
{
|
||||
uint3 gridPos;
|
||||
gridPos.x = i & gridSizeMask.x;
|
||||
gridPos.y = (i >> gridSizeShift.y) & gridSizeMask.y;
|
||||
gridPos.z = (i >> gridSizeShift.z) & gridSizeMask.z;
|
||||
return gridPos;
|
||||
}
|
||||
|
||||
// classify voxel based on number of vertices it will generate
|
||||
// one thread per voxel
|
||||
__global__ void classifyVoxel(uint *voxelVerts,
|
||||
uint *voxelOccupied,
|
||||
uchar *volume,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
uint numVoxels,
|
||||
float3 voxelSize,
|
||||
float isoValue,
|
||||
cudaTextureObject_t numVertsTex,
|
||||
cudaTextureObject_t volumeTex)
|
||||
{
|
||||
uint blockId = __mul24(blockIdx.y, gridDim.x) + blockIdx.x;
|
||||
uint i = __mul24(blockId, blockDim.x) + threadIdx.x;
|
||||
|
||||
uint3 gridPos = calcGridPos(i, gridSizeShift, gridSizeMask);
|
||||
|
||||
// read field values at neighbouring grid vertices
|
||||
#if SAMPLE_VOLUME
|
||||
float field[8];
|
||||
field[0] = sampleVolume(volumeTex, volume, gridPos, gridSize);
|
||||
field[1] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 0, 0), gridSize);
|
||||
field[2] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 1, 0), gridSize);
|
||||
field[3] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 1, 0), gridSize);
|
||||
field[4] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 0, 1), gridSize);
|
||||
field[5] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 0, 1), gridSize);
|
||||
field[6] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 1, 1), gridSize);
|
||||
field[7] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 1, 1), gridSize);
|
||||
#else
|
||||
float3 p;
|
||||
p.x = -1.0f + (gridPos.x * voxelSize.x);
|
||||
p.y = -1.0f + (gridPos.y * voxelSize.y);
|
||||
p.z = -1.0f + (gridPos.z * voxelSize.z);
|
||||
|
||||
float field[8];
|
||||
field[0] = fieldFunc(p);
|
||||
field[1] = fieldFunc(p + make_float3(voxelSize.x, 0, 0));
|
||||
field[2] = fieldFunc(p + make_float3(voxelSize.x, voxelSize.y, 0));
|
||||
field[3] = fieldFunc(p + make_float3(0, voxelSize.y, 0));
|
||||
field[4] = fieldFunc(p + make_float3(0, 0, voxelSize.z));
|
||||
field[5] = fieldFunc(p + make_float3(voxelSize.x, 0, voxelSize.z));
|
||||
field[6] = fieldFunc(p + make_float3(voxelSize.x, voxelSize.y, voxelSize.z));
|
||||
field[7] = fieldFunc(p + make_float3(0, voxelSize.y, voxelSize.z));
|
||||
#endif
|
||||
|
||||
// calculate flag indicating if each vertex is inside or outside isosurface
|
||||
uint cubeindex;
|
||||
cubeindex = uint(field[0] < isoValue);
|
||||
cubeindex += uint(field[1] < isoValue) * 2;
|
||||
cubeindex += uint(field[2] < isoValue) * 4;
|
||||
cubeindex += uint(field[3] < isoValue) * 8;
|
||||
cubeindex += uint(field[4] < isoValue) * 16;
|
||||
cubeindex += uint(field[5] < isoValue) * 32;
|
||||
cubeindex += uint(field[6] < isoValue) * 64;
|
||||
cubeindex += uint(field[7] < isoValue) * 128;
|
||||
|
||||
// read number of vertices from texture
|
||||
uint numVerts = tex1Dfetch<uint>(numVertsTex, cubeindex);
|
||||
|
||||
if (i < numVoxels) {
|
||||
voxelVerts[i] = numVerts;
|
||||
voxelOccupied[i] = (numVerts > 0);
|
||||
}
|
||||
}
|
||||
|
||||
extern "C" void launch_classifyVoxel(dim3 grid,
|
||||
dim3 threads,
|
||||
uint *voxelVerts,
|
||||
uint *voxelOccupied,
|
||||
uchar *volume,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
uint numVoxels,
|
||||
float3 voxelSize,
|
||||
float isoValue)
|
||||
{
|
||||
// calculate number of vertices need per voxel
|
||||
classifyVoxel<<<grid, threads>>>(voxelVerts,
|
||||
voxelOccupied,
|
||||
volume,
|
||||
gridSize,
|
||||
gridSizeShift,
|
||||
gridSizeMask,
|
||||
numVoxels,
|
||||
voxelSize,
|
||||
isoValue,
|
||||
numVertsTex,
|
||||
volumeTex);
|
||||
getLastCudaError("classifyVoxel failed");
|
||||
}
|
||||
|
||||
// compact voxel array
|
||||
__global__ void compactVoxels(uint *compactedVoxelArray, uint *voxelOccupied, uint *voxelOccupiedScan, uint numVoxels)
|
||||
{
|
||||
uint blockId = __mul24(blockIdx.y, gridDim.x) + blockIdx.x;
|
||||
uint i = __mul24(blockId, blockDim.x) + threadIdx.x;
|
||||
|
||||
if (voxelOccupied[i] && (i < numVoxels)) {
|
||||
compactedVoxelArray[voxelOccupiedScan[i]] = i;
|
||||
}
|
||||
}
|
||||
|
||||
extern "C" void launch_compactVoxels(dim3 grid,
|
||||
dim3 threads,
|
||||
uint *compactedVoxelArray,
|
||||
uint *voxelOccupied,
|
||||
uint *voxelOccupiedScan,
|
||||
uint numVoxels)
|
||||
{
|
||||
compactVoxels<<<grid, threads>>>(compactedVoxelArray, voxelOccupied, voxelOccupiedScan, numVoxels);
|
||||
getLastCudaError("compactVoxels failed");
|
||||
}
|
||||
|
||||
// compute interpolated vertex along an edge
|
||||
__device__ float3 vertexInterp(float isolevel, float3 p0, float3 p1, float f0, float f1)
|
||||
{
|
||||
float t = (isolevel - f0) / (f1 - f0);
|
||||
return lerp(p0, p1, t);
|
||||
}
|
||||
|
||||
// compute interpolated vertex position and normal along an edge
|
||||
__device__ void vertexInterp2(float isolevel, float3 p0, float3 p1, float4 f0, float4 f1, float3 &p, float3 &n)
|
||||
{
|
||||
float t = (isolevel - f0.w) / (f1.w - f0.w);
|
||||
p = lerp(p0, p1, t);
|
||||
n.x = lerp(f0.x, f1.x, t);
|
||||
n.y = lerp(f0.y, f1.y, t);
|
||||
n.z = lerp(f0.z, f1.z, t);
|
||||
// n = normalize(n);
|
||||
}
|
||||
|
||||
// generate triangles for each voxel using marching cubes
|
||||
// interpolates normals from field function
|
||||
__global__ void generateTriangles(float4 *pos,
|
||||
float4 *norm,
|
||||
uint *compactedVoxelArray,
|
||||
uint *numVertsScanned,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
float3 voxelSize,
|
||||
float isoValue,
|
||||
uint activeVoxels,
|
||||
uint maxVerts,
|
||||
cudaTextureObject_t triTex,
|
||||
cudaTextureObject_t numVertsTex)
|
||||
{
|
||||
uint blockId = __mul24(blockIdx.y, gridDim.x) + blockIdx.x;
|
||||
uint i = __mul24(blockId, blockDim.x) + threadIdx.x;
|
||||
|
||||
if (i > activeVoxels - 1) {
|
||||
// can't return here because of syncthreads()
|
||||
i = activeVoxels - 1;
|
||||
}
|
||||
|
||||
#if SKIP_EMPTY_VOXELS
|
||||
uint voxel = compactedVoxelArray[i];
|
||||
#else
|
||||
uint voxel = i;
|
||||
#endif
|
||||
|
||||
// compute position in 3d grid
|
||||
uint3 gridPos = calcGridPos(voxel, gridSizeShift, gridSizeMask);
|
||||
|
||||
float3 p;
|
||||
p.x = -1.0f + (gridPos.x * voxelSize.x);
|
||||
p.y = -1.0f + (gridPos.y * voxelSize.y);
|
||||
p.z = -1.0f + (gridPos.z * voxelSize.z);
|
||||
|
||||
// calculate cell vertex positions
|
||||
float3 v[8];
|
||||
v[0] = p;
|
||||
v[1] = p + make_float3(voxelSize.x, 0, 0);
|
||||
v[2] = p + make_float3(voxelSize.x, voxelSize.y, 0);
|
||||
v[3] = p + make_float3(0, voxelSize.y, 0);
|
||||
v[4] = p + make_float3(0, 0, voxelSize.z);
|
||||
v[5] = p + make_float3(voxelSize.x, 0, voxelSize.z);
|
||||
v[6] = p + make_float3(voxelSize.x, voxelSize.y, voxelSize.z);
|
||||
v[7] = p + make_float3(0, voxelSize.y, voxelSize.z);
|
||||
|
||||
// evaluate field values
|
||||
float4 field[8];
|
||||
field[0] = fieldFunc4(v[0]);
|
||||
field[1] = fieldFunc4(v[1]);
|
||||
field[2] = fieldFunc4(v[2]);
|
||||
field[3] = fieldFunc4(v[3]);
|
||||
field[4] = fieldFunc4(v[4]);
|
||||
field[5] = fieldFunc4(v[5]);
|
||||
field[6] = fieldFunc4(v[6]);
|
||||
field[7] = fieldFunc4(v[7]);
|
||||
|
||||
// recalculate flag
|
||||
// (this is faster than storing it in global memory)
|
||||
uint cubeindex;
|
||||
cubeindex = uint(field[0].w < isoValue);
|
||||
cubeindex += uint(field[1].w < isoValue) * 2;
|
||||
cubeindex += uint(field[2].w < isoValue) * 4;
|
||||
cubeindex += uint(field[3].w < isoValue) * 8;
|
||||
cubeindex += uint(field[4].w < isoValue) * 16;
|
||||
cubeindex += uint(field[5].w < isoValue) * 32;
|
||||
cubeindex += uint(field[6].w < isoValue) * 64;
|
||||
cubeindex += uint(field[7].w < isoValue) * 128;
|
||||
|
||||
// find the vertices where the surface intersects the cube
|
||||
|
||||
#if USE_SHARED
|
||||
// use partioned shared memory to avoid using local memory
|
||||
__shared__ float3 vertlist[12 * NTHREADS];
|
||||
__shared__ float3 normlist[12 * NTHREADS];
|
||||
|
||||
vertexInterp2(isoValue, v[0], v[1], field[0], field[1], vertlist[threadIdx.x], normlist[threadIdx.x]);
|
||||
vertexInterp2(
|
||||
isoValue, v[1], v[2], field[1], field[2], vertlist[threadIdx.x + NTHREADS], normlist[threadIdx.x + NTHREADS]);
|
||||
vertexInterp2(isoValue,
|
||||
v[2],
|
||||
v[3],
|
||||
field[2],
|
||||
field[3],
|
||||
vertlist[threadIdx.x + (NTHREADS * 2)],
|
||||
normlist[threadIdx.x + (NTHREADS * 2)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[3],
|
||||
v[0],
|
||||
field[3],
|
||||
field[0],
|
||||
vertlist[threadIdx.x + (NTHREADS * 3)],
|
||||
normlist[threadIdx.x + (NTHREADS * 3)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[4],
|
||||
v[5],
|
||||
field[4],
|
||||
field[5],
|
||||
vertlist[threadIdx.x + (NTHREADS * 4)],
|
||||
normlist[threadIdx.x + (NTHREADS * 4)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[5],
|
||||
v[6],
|
||||
field[5],
|
||||
field[6],
|
||||
vertlist[threadIdx.x + (NTHREADS * 5)],
|
||||
normlist[threadIdx.x + (NTHREADS * 5)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[6],
|
||||
v[7],
|
||||
field[6],
|
||||
field[7],
|
||||
vertlist[threadIdx.x + (NTHREADS * 6)],
|
||||
normlist[threadIdx.x + (NTHREADS * 6)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[7],
|
||||
v[4],
|
||||
field[7],
|
||||
field[4],
|
||||
vertlist[threadIdx.x + (NTHREADS * 7)],
|
||||
normlist[threadIdx.x + (NTHREADS * 7)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[0],
|
||||
v[4],
|
||||
field[0],
|
||||
field[4],
|
||||
vertlist[threadIdx.x + (NTHREADS * 8)],
|
||||
normlist[threadIdx.x + (NTHREADS * 8)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[1],
|
||||
v[5],
|
||||
field[1],
|
||||
field[5],
|
||||
vertlist[threadIdx.x + (NTHREADS * 9)],
|
||||
normlist[threadIdx.x + (NTHREADS * 9)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[2],
|
||||
v[6],
|
||||
field[2],
|
||||
field[6],
|
||||
vertlist[threadIdx.x + (NTHREADS * 10)],
|
||||
normlist[threadIdx.x + (NTHREADS * 10)]);
|
||||
vertexInterp2(isoValue,
|
||||
v[3],
|
||||
v[7],
|
||||
field[3],
|
||||
field[7],
|
||||
vertlist[threadIdx.x + (NTHREADS * 11)],
|
||||
normlist[threadIdx.x + (NTHREADS * 11)]);
|
||||
__syncthreads();
|
||||
|
||||
#else
|
||||
float3 vertlist[12];
|
||||
float3 normlist[12];
|
||||
|
||||
vertexInterp2(isoValue, v[0], v[1], field[0], field[1], vertlist[0], normlist[0]);
|
||||
vertexInterp2(isoValue, v[1], v[2], field[1], field[2], vertlist[1], normlist[1]);
|
||||
vertexInterp2(isoValue, v[2], v[3], field[2], field[3], vertlist[2], normlist[2]);
|
||||
vertexInterp2(isoValue, v[3], v[0], field[3], field[0], vertlist[3], normlist[3]);
|
||||
|
||||
vertexInterp2(isoValue, v[4], v[5], field[4], field[5], vertlist[4], normlist[4]);
|
||||
vertexInterp2(isoValue, v[5], v[6], field[5], field[6], vertlist[5], normlist[5]);
|
||||
vertexInterp2(isoValue, v[6], v[7], field[6], field[7], vertlist[6], normlist[6]);
|
||||
vertexInterp2(isoValue, v[7], v[4], field[7], field[4], vertlist[7], normlist[7]);
|
||||
|
||||
vertexInterp2(isoValue, v[0], v[4], field[0], field[4], vertlist[8], normlist[8]);
|
||||
vertexInterp2(isoValue, v[1], v[5], field[1], field[5], vertlist[9], normlist[9]);
|
||||
vertexInterp2(isoValue, v[2], v[6], field[2], field[6], vertlist[10], normlist[10]);
|
||||
vertexInterp2(isoValue, v[3], v[7], field[3], field[7], vertlist[11], normlist[11]);
|
||||
#endif
|
||||
|
||||
// output triangle vertices
|
||||
uint numVerts = tex1Dfetch<uint>(numVertsTex, cubeindex);
|
||||
|
||||
for (int i = 0; i < numVerts; i++) {
|
||||
uint edge = tex1Dfetch<uint>(triTex, cubeindex * 16 + i);
|
||||
|
||||
uint index = numVertsScanned[voxel] + i;
|
||||
|
||||
if (index < maxVerts) {
|
||||
#if USE_SHARED
|
||||
pos[index] = make_float4(vertlist[(edge * NTHREADS) + threadIdx.x], 1.0f);
|
||||
norm[index] = make_float4(normlist[(edge * NTHREADS) + threadIdx.x], 0.0f);
|
||||
#else
|
||||
pos[index] = make_float4(vertlist[edge], 1.0f);
|
||||
norm[index] = make_float4(normlist[edge], 0.0f);
|
||||
#endif
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
extern "C" void launch_generateTriangles(dim3 grid,
|
||||
dim3 threads,
|
||||
float4 *pos,
|
||||
float4 *norm,
|
||||
uint *compactedVoxelArray,
|
||||
uint *numVertsScanned,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
float3 voxelSize,
|
||||
float isoValue,
|
||||
uint activeVoxels,
|
||||
uint maxVerts)
|
||||
{
|
||||
generateTriangles<<<grid, NTHREADS>>>(pos,
|
||||
norm,
|
||||
compactedVoxelArray,
|
||||
numVertsScanned,
|
||||
gridSize,
|
||||
gridSizeShift,
|
||||
gridSizeMask,
|
||||
voxelSize,
|
||||
isoValue,
|
||||
activeVoxels,
|
||||
maxVerts,
|
||||
triTex,
|
||||
numVertsTex);
|
||||
getLastCudaError("generateTriangles failed");
|
||||
}
|
||||
|
||||
// calculate triangle normal
|
||||
__device__ float3 calcNormal(float3 *v0, float3 *v1, float3 *v2)
|
||||
{
|
||||
float3 edge0 = *v1 - *v0;
|
||||
float3 edge1 = *v2 - *v0;
|
||||
// note - it's faster to perform normalization in vertex shader rather than
|
||||
// here
|
||||
return cross(edge0, edge1);
|
||||
}
|
||||
|
||||
// version that calculates flat surface normal for each triangle
|
||||
__global__ void generateTriangles2(float4 *pos,
|
||||
float4 *norm,
|
||||
uint *compactedVoxelArray,
|
||||
uint *numVertsScanned,
|
||||
uchar *volume,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
float3 voxelSize,
|
||||
float isoValue,
|
||||
uint activeVoxels,
|
||||
uint maxVerts,
|
||||
cudaTextureObject_t triTex,
|
||||
cudaTextureObject_t numVertsTex,
|
||||
cudaTextureObject_t volumeTex)
|
||||
{
|
||||
uint blockId = __mul24(blockIdx.y, gridDim.x) + blockIdx.x;
|
||||
uint i = __mul24(blockId, blockDim.x) + threadIdx.x;
|
||||
|
||||
if (i > activeVoxels - 1) {
|
||||
i = activeVoxels - 1;
|
||||
}
|
||||
|
||||
#if SKIP_EMPTY_VOXELS
|
||||
uint voxel = compactedVoxelArray[i];
|
||||
#else
|
||||
uint voxel = i;
|
||||
#endif
|
||||
|
||||
// compute position in 3d grid
|
||||
uint3 gridPos = calcGridPos(voxel, gridSizeShift, gridSizeMask);
|
||||
|
||||
float3 p;
|
||||
p.x = -1.0f + (gridPos.x * voxelSize.x);
|
||||
p.y = -1.0f + (gridPos.y * voxelSize.y);
|
||||
p.z = -1.0f + (gridPos.z * voxelSize.z);
|
||||
|
||||
// calculate cell vertex positions
|
||||
float3 v[8];
|
||||
v[0] = p;
|
||||
v[1] = p + make_float3(voxelSize.x, 0, 0);
|
||||
v[2] = p + make_float3(voxelSize.x, voxelSize.y, 0);
|
||||
v[3] = p + make_float3(0, voxelSize.y, 0);
|
||||
v[4] = p + make_float3(0, 0, voxelSize.z);
|
||||
v[5] = p + make_float3(voxelSize.x, 0, voxelSize.z);
|
||||
v[6] = p + make_float3(voxelSize.x, voxelSize.y, voxelSize.z);
|
||||
v[7] = p + make_float3(0, voxelSize.y, voxelSize.z);
|
||||
|
||||
#if SAMPLE_VOLUME
|
||||
float field[8];
|
||||
field[0] = sampleVolume(volumeTex, volume, gridPos, gridSize);
|
||||
field[1] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 0, 0), gridSize);
|
||||
field[2] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 1, 0), gridSize);
|
||||
field[3] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 1, 0), gridSize);
|
||||
field[4] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 0, 1), gridSize);
|
||||
field[5] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 0, 1), gridSize);
|
||||
field[6] = sampleVolume(volumeTex, volume, gridPos + make_uint3(1, 1, 1), gridSize);
|
||||
field[7] = sampleVolume(volumeTex, volume, gridPos + make_uint3(0, 1, 1), gridSize);
|
||||
#else
|
||||
// evaluate field values
|
||||
float field[8];
|
||||
field[0] = fieldFunc(v[0]);
|
||||
field[1] = fieldFunc(v[1]);
|
||||
field[2] = fieldFunc(v[2]);
|
||||
field[3] = fieldFunc(v[3]);
|
||||
field[4] = fieldFunc(v[4]);
|
||||
field[5] = fieldFunc(v[5]);
|
||||
field[6] = fieldFunc(v[6]);
|
||||
field[7] = fieldFunc(v[7]);
|
||||
#endif
|
||||
|
||||
// recalculate flag
|
||||
uint cubeindex;
|
||||
cubeindex = uint(field[0] < isoValue);
|
||||
cubeindex += uint(field[1] < isoValue) * 2;
|
||||
cubeindex += uint(field[2] < isoValue) * 4;
|
||||
cubeindex += uint(field[3] < isoValue) * 8;
|
||||
cubeindex += uint(field[4] < isoValue) * 16;
|
||||
cubeindex += uint(field[5] < isoValue) * 32;
|
||||
cubeindex += uint(field[6] < isoValue) * 64;
|
||||
cubeindex += uint(field[7] < isoValue) * 128;
|
||||
|
||||
// find the vertices where the surface intersects the cube
|
||||
|
||||
#if USE_SHARED
|
||||
// use shared memory to avoid using local
|
||||
__shared__ float3 vertlist[12 * NTHREADS];
|
||||
|
||||
vertlist[threadIdx.x] = vertexInterp(isoValue, v[0], v[1], field[0], field[1]);
|
||||
vertlist[NTHREADS + threadIdx.x] = vertexInterp(isoValue, v[1], v[2], field[1], field[2]);
|
||||
vertlist[(NTHREADS * 2) + threadIdx.x] = vertexInterp(isoValue, v[2], v[3], field[2], field[3]);
|
||||
vertlist[(NTHREADS * 3) + threadIdx.x] = vertexInterp(isoValue, v[3], v[0], field[3], field[0]);
|
||||
vertlist[(NTHREADS * 4) + threadIdx.x] = vertexInterp(isoValue, v[4], v[5], field[4], field[5]);
|
||||
vertlist[(NTHREADS * 5) + threadIdx.x] = vertexInterp(isoValue, v[5], v[6], field[5], field[6]);
|
||||
vertlist[(NTHREADS * 6) + threadIdx.x] = vertexInterp(isoValue, v[6], v[7], field[6], field[7]);
|
||||
vertlist[(NTHREADS * 7) + threadIdx.x] = vertexInterp(isoValue, v[7], v[4], field[7], field[4]);
|
||||
vertlist[(NTHREADS * 8) + threadIdx.x] = vertexInterp(isoValue, v[0], v[4], field[0], field[4]);
|
||||
vertlist[(NTHREADS * 9) + threadIdx.x] = vertexInterp(isoValue, v[1], v[5], field[1], field[5]);
|
||||
vertlist[(NTHREADS * 10) + threadIdx.x] = vertexInterp(isoValue, v[2], v[6], field[2], field[6]);
|
||||
vertlist[(NTHREADS * 11) + threadIdx.x] = vertexInterp(isoValue, v[3], v[7], field[3], field[7]);
|
||||
__syncthreads();
|
||||
#else
|
||||
|
||||
float3 vertlist[12];
|
||||
|
||||
vertlist[0] = vertexInterp(isoValue, v[0], v[1], field[0], field[1]);
|
||||
vertlist[1] = vertexInterp(isoValue, v[1], v[2], field[1], field[2]);
|
||||
vertlist[2] = vertexInterp(isoValue, v[2], v[3], field[2], field[3]);
|
||||
vertlist[3] = vertexInterp(isoValue, v[3], v[0], field[3], field[0]);
|
||||
|
||||
vertlist[4] = vertexInterp(isoValue, v[4], v[5], field[4], field[5]);
|
||||
vertlist[5] = vertexInterp(isoValue, v[5], v[6], field[5], field[6]);
|
||||
vertlist[6] = vertexInterp(isoValue, v[6], v[7], field[6], field[7]);
|
||||
vertlist[7] = vertexInterp(isoValue, v[7], v[4], field[7], field[4]);
|
||||
|
||||
vertlist[8] = vertexInterp(isoValue, v[0], v[4], field[0], field[4]);
|
||||
vertlist[9] = vertexInterp(isoValue, v[1], v[5], field[1], field[5]);
|
||||
vertlist[10] = vertexInterp(isoValue, v[2], v[6], field[2], field[6]);
|
||||
vertlist[11] = vertexInterp(isoValue, v[3], v[7], field[3], field[7]);
|
||||
#endif
|
||||
|
||||
// output triangle vertices
|
||||
uint numVerts = tex1Dfetch<uint>(numVertsTex, cubeindex);
|
||||
|
||||
for (int i = 0; i < numVerts; i += 3) {
|
||||
uint index = numVertsScanned[voxel] + i;
|
||||
|
||||
float3 *v[3];
|
||||
uint edge;
|
||||
edge = tex1Dfetch<uint>(triTex, (cubeindex * 16) + i);
|
||||
#if USE_SHARED
|
||||
v[0] = &vertlist[(edge * NTHREADS) + threadIdx.x];
|
||||
#else
|
||||
v[0] = &vertlist[edge];
|
||||
#endif
|
||||
|
||||
edge = tex1Dfetch<uint>(triTex, (cubeindex * 16) + i + 1);
|
||||
#if USE_SHARED
|
||||
v[1] = &vertlist[(edge * NTHREADS) + threadIdx.x];
|
||||
#else
|
||||
v[1] = &vertlist[edge];
|
||||
#endif
|
||||
|
||||
edge = tex1Dfetch<uint>(triTex, (cubeindex * 16) + i + 2);
|
||||
#if USE_SHARED
|
||||
v[2] = &vertlist[(edge * NTHREADS) + threadIdx.x];
|
||||
#else
|
||||
v[2] = &vertlist[edge];
|
||||
#endif
|
||||
|
||||
// calculate triangle surface normal
|
||||
float3 n = calcNormal(v[0], v[1], v[2]);
|
||||
|
||||
if (index < (maxVerts - 3)) {
|
||||
pos[index] = make_float4(*v[0], 1.0f);
|
||||
norm[index] = make_float4(n, 0.0f);
|
||||
|
||||
pos[index + 1] = make_float4(*v[1], 1.0f);
|
||||
norm[index + 1] = make_float4(n, 0.0f);
|
||||
|
||||
pos[index + 2] = make_float4(*v[2], 1.0f);
|
||||
norm[index + 2] = make_float4(n, 0.0f);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
extern "C" void launch_generateTriangles2(dim3 grid,
|
||||
dim3 threads,
|
||||
float4 *pos,
|
||||
float4 *norm,
|
||||
uint *compactedVoxelArray,
|
||||
uint *numVertsScanned,
|
||||
uchar *volume,
|
||||
uint3 gridSize,
|
||||
uint3 gridSizeShift,
|
||||
uint3 gridSizeMask,
|
||||
float3 voxelSize,
|
||||
float isoValue,
|
||||
uint activeVoxels,
|
||||
uint maxVerts)
|
||||
{
|
||||
generateTriangles2<<<grid, NTHREADS>>>(pos,
|
||||
norm,
|
||||
compactedVoxelArray,
|
||||
numVertsScanned,
|
||||
volume,
|
||||
gridSize,
|
||||
gridSizeShift,
|
||||
gridSizeMask,
|
||||
voxelSize,
|
||||
isoValue,
|
||||
activeVoxels,
|
||||
maxVerts,
|
||||
triTex,
|
||||
numVertsTex,
|
||||
volumeTex);
|
||||
getLastCudaError("generateTriangles2 failed");
|
||||
}
|
||||
|
||||
extern "C" void ThrustScanWrapper(unsigned int *output, unsigned int *input, unsigned int numElements)
|
||||
{
|
||||
thrust::exclusive_scan(thrust::device_ptr<unsigned int>(input),
|
||||
thrust::device_ptr<unsigned int>(input + numElements),
|
||||
thrust::device_ptr<unsigned int>(output));
|
||||
}
|
||||
|
||||
#endif
|
||||
198
cpp/5_Domain_Specific/marchingCubes/tables.h
Normal file
198
cpp/5_Domain_Specific/marchingCubes/tables.h
Normal file
@@ -0,0 +1,198 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
/*
|
||||
Tables for Marching Cubes
|
||||
http://local.wasp.uwa.edu.au/~pbourke/geometry/polygonise/
|
||||
*/
|
||||
|
||||
// edge table maps 8-bit flag representing which cube vertices are inside
|
||||
// the isosurface to 12-bit number indicating which edges are intersected
|
||||
uint edgeTable[256] = {
|
||||
0x0, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
|
||||
0x190, 0x99, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
|
||||
0x230, 0x339, 0x33, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
|
||||
0x3a0, 0x2a9, 0x1a3, 0xaa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
|
||||
0x460, 0x569, 0x663, 0x76a, 0x66, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
|
||||
0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
|
||||
0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
|
||||
0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
|
||||
0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0xcc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
|
||||
0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x55, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
|
||||
0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0xff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
|
||||
0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x66, 0x76a, 0x663, 0x569, 0x460,
|
||||
0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa, 0x1a3, 0x2a9, 0x3a0,
|
||||
0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33, 0x339, 0x230,
|
||||
0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99, 0x190,
|
||||
0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0};
|
||||
|
||||
// triangle table maps same cube vertex index to a list of up to 5 triangles
|
||||
// which are built from the interpolated edge vertices
|
||||
#define X 255
|
||||
uint triTable[256][16] = {
|
||||
{X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X}, {0, 8, 3, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{0, 1, 9, X, X, X, X, X, X, X, X, X, X, X, X, X}, {1, 8, 3, 9, 8, 1, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 2, 10, X, X, X, X, X, X, X, X, X, X, X, X, X}, {0, 8, 3, 1, 2, 10, X, X, X, X, X, X, X, X, X, X},
|
||||
{9, 2, 10, 0, 2, 9, X, X, X, X, X, X, X, X, X, X}, {2, 8, 3, 2, 10, 8, 10, 9, 8, X, X, X, X, X, X, X},
|
||||
{3, 11, 2, X, X, X, X, X, X, X, X, X, X, X, X, X}, {0, 11, 2, 8, 11, 0, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 9, 0, 2, 3, 11, X, X, X, X, X, X, X, X, X, X}, {1, 11, 2, 1, 9, 11, 9, 8, 11, X, X, X, X, X, X, X},
|
||||
{3, 10, 1, 11, 10, 3, X, X, X, X, X, X, X, X, X, X}, {0, 10, 1, 0, 8, 10, 8, 11, 10, X, X, X, X, X, X, X},
|
||||
{3, 9, 0, 3, 11, 9, 11, 10, 9, X, X, X, X, X, X, X}, {9, 8, 10, 10, 8, 11, X, X, X, X, X, X, X, X, X, X},
|
||||
{4, 7, 8, X, X, X, X, X, X, X, X, X, X, X, X, X}, {4, 3, 0, 7, 3, 4, X, X, X, X, X, X, X, X, X, X},
|
||||
{0, 1, 9, 8, 4, 7, X, X, X, X, X, X, X, X, X, X}, {4, 1, 9, 4, 7, 1, 7, 3, 1, X, X, X, X, X, X, X},
|
||||
{1, 2, 10, 8, 4, 7, X, X, X, X, X, X, X, X, X, X}, {3, 4, 7, 3, 0, 4, 1, 2, 10, X, X, X, X, X, X, X},
|
||||
{9, 2, 10, 9, 0, 2, 8, 4, 7, X, X, X, X, X, X, X}, {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, X, X, X, X},
|
||||
{8, 4, 7, 3, 11, 2, X, X, X, X, X, X, X, X, X, X}, {11, 4, 7, 11, 2, 4, 2, 0, 4, X, X, X, X, X, X, X},
|
||||
{9, 0, 1, 8, 4, 7, 2, 3, 11, X, X, X, X, X, X, X}, {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, X, X, X, X},
|
||||
{3, 10, 1, 3, 11, 10, 7, 8, 4, X, X, X, X, X, X, X}, {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, X, X, X, X},
|
||||
{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, X, X, X, X}, {4, 7, 11, 4, 11, 9, 9, 11, 10, X, X, X, X, X, X, X},
|
||||
{9, 5, 4, X, X, X, X, X, X, X, X, X, X, X, X, X}, {9, 5, 4, 0, 8, 3, X, X, X, X, X, X, X, X, X, X},
|
||||
{0, 5, 4, 1, 5, 0, X, X, X, X, X, X, X, X, X, X}, {8, 5, 4, 8, 3, 5, 3, 1, 5, X, X, X, X, X, X, X},
|
||||
{1, 2, 10, 9, 5, 4, X, X, X, X, X, X, X, X, X, X}, {3, 0, 8, 1, 2, 10, 4, 9, 5, X, X, X, X, X, X, X},
|
||||
{5, 2, 10, 5, 4, 2, 4, 0, 2, X, X, X, X, X, X, X}, {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, X, X, X, X},
|
||||
{9, 5, 4, 2, 3, 11, X, X, X, X, X, X, X, X, X, X}, {0, 11, 2, 0, 8, 11, 4, 9, 5, X, X, X, X, X, X, X},
|
||||
{0, 5, 4, 0, 1, 5, 2, 3, 11, X, X, X, X, X, X, X}, {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, X, X, X, X},
|
||||
{10, 3, 11, 10, 1, 3, 9, 5, 4, X, X, X, X, X, X, X}, {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, X, X, X, X},
|
||||
{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, X, X, X, X}, {5, 4, 8, 5, 8, 10, 10, 8, 11, X, X, X, X, X, X, X},
|
||||
{9, 7, 8, 5, 7, 9, X, X, X, X, X, X, X, X, X, X}, {9, 3, 0, 9, 5, 3, 5, 7, 3, X, X, X, X, X, X, X},
|
||||
{0, 7, 8, 0, 1, 7, 1, 5, 7, X, X, X, X, X, X, X}, {1, 5, 3, 3, 5, 7, X, X, X, X, X, X, X, X, X, X},
|
||||
{9, 7, 8, 9, 5, 7, 10, 1, 2, X, X, X, X, X, X, X}, {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, X, X, X, X},
|
||||
{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, X, X, X, X}, {2, 10, 5, 2, 5, 3, 3, 5, 7, X, X, X, X, X, X, X},
|
||||
{7, 9, 5, 7, 8, 9, 3, 11, 2, X, X, X, X, X, X, X}, {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, X, X, X, X},
|
||||
{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, X, X, X, X}, {11, 2, 1, 11, 1, 7, 7, 1, 5, X, X, X, X, X, X, X},
|
||||
{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, X, X, X, X}, {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, X},
|
||||
{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, X}, {11, 10, 5, 7, 11, 5, X, X, X, X, X, X, X, X, X, X},
|
||||
{10, 6, 5, X, X, X, X, X, X, X, X, X, X, X, X, X}, {0, 8, 3, 5, 10, 6, X, X, X, X, X, X, X, X, X, X},
|
||||
{9, 0, 1, 5, 10, 6, X, X, X, X, X, X, X, X, X, X}, {1, 8, 3, 1, 9, 8, 5, 10, 6, X, X, X, X, X, X, X},
|
||||
{1, 6, 5, 2, 6, 1, X, X, X, X, X, X, X, X, X, X}, {1, 6, 5, 1, 2, 6, 3, 0, 8, X, X, X, X, X, X, X},
|
||||
{9, 6, 5, 9, 0, 6, 0, 2, 6, X, X, X, X, X, X, X}, {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, X, X, X, X},
|
||||
{2, 3, 11, 10, 6, 5, X, X, X, X, X, X, X, X, X, X}, {11, 0, 8, 11, 2, 0, 10, 6, 5, X, X, X, X, X, X, X},
|
||||
{0, 1, 9, 2, 3, 11, 5, 10, 6, X, X, X, X, X, X, X}, {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, X, X, X, X},
|
||||
{6, 3, 11, 6, 5, 3, 5, 1, 3, X, X, X, X, X, X, X}, {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, X, X, X, X},
|
||||
{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, X, X, X, X}, {6, 5, 9, 6, 9, 11, 11, 9, 8, X, X, X, X, X, X, X},
|
||||
{5, 10, 6, 4, 7, 8, X, X, X, X, X, X, X, X, X, X}, {4, 3, 0, 4, 7, 3, 6, 5, 10, X, X, X, X, X, X, X},
|
||||
{1, 9, 0, 5, 10, 6, 8, 4, 7, X, X, X, X, X, X, X}, {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, X, X, X, X},
|
||||
{6, 1, 2, 6, 5, 1, 4, 7, 8, X, X, X, X, X, X, X}, {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, X, X, X, X},
|
||||
{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, X, X, X, X}, {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, X},
|
||||
{3, 11, 2, 7, 8, 4, 10, 6, 5, X, X, X, X, X, X, X}, {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, X, X, X, X},
|
||||
{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, X, X, X, X}, {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, X},
|
||||
{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, X, X, X, X}, {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, X},
|
||||
{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, X}, {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, X, X, X, X},
|
||||
{10, 4, 9, 6, 4, 10, X, X, X, X, X, X, X, X, X, X}, {4, 10, 6, 4, 9, 10, 0, 8, 3, X, X, X, X, X, X, X},
|
||||
{10, 0, 1, 10, 6, 0, 6, 4, 0, X, X, X, X, X, X, X}, {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, X, X, X, X},
|
||||
{1, 4, 9, 1, 2, 4, 2, 6, 4, X, X, X, X, X, X, X}, {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, X, X, X, X},
|
||||
{0, 2, 4, 4, 2, 6, X, X, X, X, X, X, X, X, X, X}, {8, 3, 2, 8, 2, 4, 4, 2, 6, X, X, X, X, X, X, X},
|
||||
{10, 4, 9, 10, 6, 4, 11, 2, 3, X, X, X, X, X, X, X}, {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, X, X, X, X},
|
||||
{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, X, X, X, X}, {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, X},
|
||||
{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, X, X, X, X}, {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, X},
|
||||
{3, 11, 6, 3, 6, 0, 0, 6, 4, X, X, X, X, X, X, X}, {6, 4, 8, 11, 6, 8, X, X, X, X, X, X, X, X, X, X},
|
||||
{7, 10, 6, 7, 8, 10, 8, 9, 10, X, X, X, X, X, X, X}, {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, X, X, X, X},
|
||||
{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, X, X, X, X}, {10, 6, 7, 10, 7, 1, 1, 7, 3, X, X, X, X, X, X, X},
|
||||
{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, X, X, X, X}, {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, X},
|
||||
{7, 8, 0, 7, 0, 6, 6, 0, 2, X, X, X, X, X, X, X}, {7, 3, 2, 6, 7, 2, X, X, X, X, X, X, X, X, X, X},
|
||||
{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, X, X, X, X}, {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, X},
|
||||
{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, X}, {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, X, X, X, X},
|
||||
{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, X}, {0, 9, 1, 11, 6, 7, X, X, X, X, X, X, X, X, X, X},
|
||||
{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, X, X, X, X}, {7, 11, 6, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{7, 6, 11, X, X, X, X, X, X, X, X, X, X, X, X, X}, {3, 0, 8, 11, 7, 6, X, X, X, X, X, X, X, X, X, X},
|
||||
{0, 1, 9, 11, 7, 6, X, X, X, X, X, X, X, X, X, X}, {8, 1, 9, 8, 3, 1, 11, 7, 6, X, X, X, X, X, X, X},
|
||||
{10, 1, 2, 6, 11, 7, X, X, X, X, X, X, X, X, X, X}, {1, 2, 10, 3, 0, 8, 6, 11, 7, X, X, X, X, X, X, X},
|
||||
{2, 9, 0, 2, 10, 9, 6, 11, 7, X, X, X, X, X, X, X}, {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, X, X, X, X},
|
||||
{7, 2, 3, 6, 2, 7, X, X, X, X, X, X, X, X, X, X}, {7, 0, 8, 7, 6, 0, 6, 2, 0, X, X, X, X, X, X, X},
|
||||
{2, 7, 6, 2, 3, 7, 0, 1, 9, X, X, X, X, X, X, X}, {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, X, X, X, X},
|
||||
{10, 7, 6, 10, 1, 7, 1, 3, 7, X, X, X, X, X, X, X}, {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, X, X, X, X},
|
||||
{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, X, X, X, X}, {7, 6, 10, 7, 10, 8, 8, 10, 9, X, X, X, X, X, X, X},
|
||||
{6, 8, 4, 11, 8, 6, X, X, X, X, X, X, X, X, X, X}, {3, 6, 11, 3, 0, 6, 0, 4, 6, X, X, X, X, X, X, X},
|
||||
{8, 6, 11, 8, 4, 6, 9, 0, 1, X, X, X, X, X, X, X}, {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, X, X, X, X},
|
||||
{6, 8, 4, 6, 11, 8, 2, 10, 1, X, X, X, X, X, X, X}, {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, X, X, X, X},
|
||||
{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, X, X, X, X}, {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, X},
|
||||
{8, 2, 3, 8, 4, 2, 4, 6, 2, X, X, X, X, X, X, X}, {0, 4, 2, 4, 6, 2, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, X, X, X, X}, {1, 9, 4, 1, 4, 2, 2, 4, 6, X, X, X, X, X, X, X},
|
||||
{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, X, X, X, X}, {10, 1, 0, 10, 0, 6, 6, 0, 4, X, X, X, X, X, X, X},
|
||||
{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, X}, {10, 9, 4, 6, 10, 4, X, X, X, X, X, X, X, X, X, X},
|
||||
{4, 9, 5, 7, 6, 11, X, X, X, X, X, X, X, X, X, X}, {0, 8, 3, 4, 9, 5, 11, 7, 6, X, X, X, X, X, X, X},
|
||||
{5, 0, 1, 5, 4, 0, 7, 6, 11, X, X, X, X, X, X, X}, {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, X, X, X, X},
|
||||
{9, 5, 4, 10, 1, 2, 7, 6, 11, X, X, X, X, X, X, X}, {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, X, X, X, X},
|
||||
{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, X, X, X, X}, {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, X},
|
||||
{7, 2, 3, 7, 6, 2, 5, 4, 9, X, X, X, X, X, X, X}, {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, X, X, X, X},
|
||||
{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, X, X, X, X}, {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, X},
|
||||
{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, X, X, X, X}, {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, X},
|
||||
{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, X}, {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, X, X, X, X},
|
||||
{6, 9, 5, 6, 11, 9, 11, 8, 9, X, X, X, X, X, X, X}, {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, X, X, X, X},
|
||||
{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, X, X, X, X}, {6, 11, 3, 6, 3, 5, 5, 3, 1, X, X, X, X, X, X, X},
|
||||
{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, X, X, X, X}, {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, X},
|
||||
{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, X}, {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, X, X, X, X},
|
||||
{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, X, X, X, X}, {9, 5, 6, 9, 6, 0, 0, 6, 2, X, X, X, X, X, X, X},
|
||||
{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, X}, {1, 5, 6, 2, 1, 6, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, X}, {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, X, X, X, X},
|
||||
{0, 3, 8, 5, 6, 10, X, X, X, X, X, X, X, X, X, X}, {10, 5, 6, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{11, 5, 10, 7, 5, 11, X, X, X, X, X, X, X, X, X, X}, {11, 5, 10, 11, 7, 5, 8, 3, 0, X, X, X, X, X, X, X},
|
||||
{5, 11, 7, 5, 10, 11, 1, 9, 0, X, X, X, X, X, X, X}, {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, X, X, X, X},
|
||||
{11, 1, 2, 11, 7, 1, 7, 5, 1, X, X, X, X, X, X, X}, {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, X, X, X, X},
|
||||
{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, X, X, X, X}, {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, X},
|
||||
{2, 5, 10, 2, 3, 5, 3, 7, 5, X, X, X, X, X, X, X}, {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, X, X, X, X},
|
||||
{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, X, X, X, X}, {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, X},
|
||||
{1, 3, 5, 3, 7, 5, X, X, X, X, X, X, X, X, X, X}, {0, 8, 7, 0, 7, 1, 1, 7, 5, X, X, X, X, X, X, X},
|
||||
{9, 0, 3, 9, 3, 5, 5, 3, 7, X, X, X, X, X, X, X}, {9, 8, 7, 5, 9, 7, X, X, X, X, X, X, X, X, X, X},
|
||||
{5, 8, 4, 5, 10, 8, 10, 11, 8, X, X, X, X, X, X, X}, {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, X, X, X, X},
|
||||
{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, X, X, X, X}, {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, X},
|
||||
{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, X, X, X, X}, {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, X},
|
||||
{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, X}, {9, 4, 5, 2, 11, 3, X, X, X, X, X, X, X, X, X, X},
|
||||
{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, X, X, X, X}, {5, 10, 2, 5, 2, 4, 4, 2, 0, X, X, X, X, X, X, X},
|
||||
{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, X}, {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, X, X, X, X},
|
||||
{8, 4, 5, 8, 5, 3, 3, 5, 1, X, X, X, X, X, X, X}, {0, 4, 5, 1, 0, 5, X, X, X, X, X, X, X, X, X, X},
|
||||
{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, X, X, X, X}, {9, 4, 5, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{4, 11, 7, 4, 9, 11, 9, 10, 11, X, X, X, X, X, X, X}, {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, X, X, X, X},
|
||||
{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, X, X, X, X}, {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, X},
|
||||
{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, X, X, X, X}, {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, X},
|
||||
{11, 7, 4, 11, 4, 2, 2, 4, 0, X, X, X, X, X, X, X}, {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, X, X, X, X},
|
||||
{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, X, X, X, X}, {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, X},
|
||||
{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, X}, {1, 10, 2, 8, 7, 4, X, X, X, X, X, X, X, X, X, X},
|
||||
{4, 9, 1, 4, 1, 7, 7, 1, 3, X, X, X, X, X, X, X}, {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, X, X, X, X},
|
||||
{4, 0, 3, 7, 4, 3, X, X, X, X, X, X, X, X, X, X}, {4, 8, 7, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{9, 10, 8, 10, 11, 8, X, X, X, X, X, X, X, X, X, X}, {3, 0, 9, 3, 9, 11, 11, 9, 10, X, X, X, X, X, X, X},
|
||||
{0, 1, 10, 0, 10, 8, 8, 10, 11, X, X, X, X, X, X, X}, {3, 1, 10, 11, 3, 10, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 2, 11, 1, 11, 9, 9, 11, 8, X, X, X, X, X, X, X}, {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, X, X, X, X},
|
||||
{0, 2, 11, 8, 0, 11, X, X, X, X, X, X, X, X, X, X}, {3, 2, 11, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{2, 3, 8, 2, 8, 10, 10, 8, 9, X, X, X, X, X, X, X}, {9, 10, 2, 0, 9, 2, X, X, X, X, X, X, X, X, X, X},
|
||||
{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, X, X, X, X}, {1, 10, 2, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{1, 3, 8, 9, 1, 8, X, X, X, X, X, X, X, X, X, X}, {0, 9, 1, X, X, X, X, X, X, X, X, X, X, X, X, X},
|
||||
{0, 3, 8, X, X, X, X, X, X, X, X, X, X, X, X, X}, {X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X}};
|
||||
#undef X
|
||||
|
||||
// number of vertices for each case above
|
||||
uint numVertsTable[256] = {
|
||||
0, 3, 3, 6, 3, 6, 6, 9, 3, 6, 6, 9, 6, 9, 9, 6, 3, 6, 6, 9, 6, 9, 9, 12, 6, 9, 9, 12, 9,
|
||||
12, 12, 9, 3, 6, 6, 9, 6, 9, 9, 12, 6, 9, 9, 12, 9, 12, 12, 9, 6, 9, 9, 6, 9, 12, 12, 9, 9, 12,
|
||||
12, 9, 12, 15, 15, 6, 3, 6, 6, 9, 6, 9, 9, 12, 6, 9, 9, 12, 9, 12, 12, 9, 6, 9, 9, 12, 9, 12, 12,
|
||||
15, 9, 12, 12, 15, 12, 15, 15, 12, 6, 9, 9, 12, 9, 12, 6, 9, 9, 12, 12, 15, 12, 15, 9, 6, 9, 12, 12, 9,
|
||||
12, 15, 9, 6, 12, 15, 15, 12, 15, 6, 12, 3, 3, 6, 6, 9, 6, 9, 9, 12, 6, 9, 9, 12, 9, 12, 12, 9, 6,
|
||||
9, 9, 12, 9, 12, 12, 15, 9, 6, 12, 9, 12, 9, 15, 6, 6, 9, 9, 12, 9, 12, 12, 15, 9, 12, 12, 15, 12, 15,
|
||||
15, 12, 9, 12, 12, 9, 12, 15, 15, 12, 12, 9, 15, 6, 15, 12, 6, 3, 6, 9, 9, 12, 9, 12, 12, 15, 9, 12, 12,
|
||||
15, 6, 9, 9, 6, 9, 12, 12, 15, 12, 15, 15, 6, 12, 9, 15, 12, 9, 6, 12, 3, 9, 12, 12, 15, 12, 15, 9, 12,
|
||||
12, 15, 15, 6, 9, 12, 6, 3, 6, 9, 9, 6, 9, 12, 6, 3, 9, 6, 12, 3, 6, 3, 3, 0,
|
||||
};
|
||||
Reference in New Issue
Block a user